﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Net;
using System.Text;
using System.Threading.Tasks;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;

namespace VitesseClientDemo
{
    class AppController : INotifyPropertyChanged
    {
        public MainWindow Window { get; private set; }

        private const string txtConnect = "Connect";
        private const string txtDisconnect = "Disconnect";
        private string connectAction = txtConnect;

        private string serverIp = "";
        private int serverPort = 19183;

        private AsyncClient Client { get; set; } = null;

        private string rawBuf = "";
        private string jsonBuf = "";
        private string compBuf = "";
        private string eventsBuf = "";

        private bool msgInfo = true;
        private bool msgSkaters = true;
        private bool msgRaces = true;

        private Json.JsonMessages MessageParser { get; set; }

        public AppController(MainWindow window)
        {
            Window = window;
            Window.DataContext = this;

            serverIp = Properties.Settings.Default.ServerIp;
            serverPort = Properties.Settings.Default.ServerPort;

            MessageParser = new Json.JsonMessages((s) => { CompBuf += s; }, (s) => { EventsBuf += s + "\n"; });

            CommandsBind();
        }

        #region Server Info
        /// <summary>
        /// Get/set the server address
        /// </summary>
        public string ServerIp
        {
            get { return serverIp; }
            set
            {
                IPAddress newAddr;
                if (IPAddress.TryParse(value, out newAddr)) {
                    serverIp = value;
                    Properties.Settings.Default.ServerIp = serverIp;
                    Properties.Settings.Default.Save();
                }

                OnPropertyChanged();
            }
        }

        /// <summary>
        /// Get/set the server port
        /// </summary>
        public string ServerPort
        {
            get { return serverPort.ToString(); }
            set
            {
                int newPort;
                if (int.TryParse(value, out newPort))
                {
                    if ((newPort > 0) && (newPort < 65536))
                    {
                        serverPort = newPort;
                        Properties.Settings.Default.ServerPort = serverPort;
                        Properties.Settings.Default.Save();
                    }
                }

                OnPropertyChanged();
            }
        }

        /// <summary>
        /// Get/set whether info messages are wanted
        /// </summary>
        public bool MsgInfo
        {
            get { return msgInfo; }
            set
            {
                if (SetField(ref msgInfo, value))
                {
                    ClientSend(MessageParser.MsgOptionsBuild(MsgInfo, MsgSkaters, MsgRaces));
                }
            }
        }

        /// <summary>
        /// Get/set whether skater messages are wanted
        /// </summary>
        public bool MsgSkaters
        {
            get { return msgSkaters; }
            set
            {
                if (SetField(ref msgSkaters, value))
                {
                    ClientSend(MessageParser.MsgOptionsBuild(MsgInfo, MsgSkaters, MsgRaces));
                }
            }
        }

        /// <summary>
        /// Get/set whether race messages are wanted
        /// </summary>
        public bool MsgRaces
        {
            get { return msgRaces; }
            set
            {
                if (SetField(ref msgRaces, value))
                {
                    ClientSend(MessageParser.MsgOptionsBuild(MsgInfo, MsgSkaters, MsgRaces));
                }
            }
        }
        #endregion

        /// <summary>
        /// Text for connect button
        /// </summary>
        public string ConnectAction
        {
            get { return connectAction; }
            set { SetField(ref connectAction, value); }
        }

        public string RawBuf
        {
            get { return rawBuf; }
            set { SetField(ref rawBuf, value); }
        }

        public string JSONBuf
        {
            get { return jsonBuf; }
            set { SetField(ref jsonBuf, value); }
        }

        public string CompBuf
        {
            get { return compBuf; }
            set { SetField(ref compBuf, value); }
        }

        public string EventsBuf
        {
            get { return eventsBuf; }
            set { SetField(ref eventsBuf, value); }
        }

        private void BuffersClear()
        {
            RawBuf = "";
            JSONBuf = "";
            CompBuf = "";
        }

        #region Client
        /// <summary>
        /// Connect to the server
        /// </summary>
        private void ClientConnect()
        {
            BuffersClear();

            Client = new AsyncClient(serverIp, serverPort);

            Client.ServerConnected += Client_ServerConnected;
            Client.ServerDisconnected += Client_ServerDisconnected;
            Client.MessageReceived += Client_MessageReceived;
            Client.Start();
        }

        /// <summary>
        /// Disconnect from the server
        /// </summary>
        private void ClientDisconnect()
        {
            if (Client != null)
            {
                Client.Dispose();
                Client = null;
            }
        }

        /// <summary>
        /// Handle a received message
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="msg"></param>
        private void Client_MessageReceived(object sender, string msg)
        {
            RawBuf += msg + "\n";

            try {
                JObject json = JObject.Parse(msg);
                JSONBuf += json.ToString() + "\n";

                MessageParser.Parse(json);
            } catch
            {
                LogEvent("JSON not valid");
            }
        }

        /// <summary>
        /// Server disconnected notification
        /// </summary>
        /// <param name="sender"></param>
        private void Client_ServerDisconnected(object sender)
        {
            LogEvent("Disconnected");
            Client = null;
            ConnectAction = txtConnect;
        }

        /// <summary>
        /// Server connected notification
        /// </summary>
        /// <param name="sender"></param>
        private void Client_ServerConnected(object sender)
        {
            LogEvent("Connected");
            ConnectAction = txtDisconnect;
            ClientSend(MessageParser.MsgOptionsBuild(MsgInfo, MsgSkaters, MsgRaces));
        }

        /// <summary>
        /// Send message to the server
        /// </summary>
        /// <param name="msg"></param>
        private void ClientSend(string msg)
        {
            if (Client != null)
            {
                Client.Send(msg);
            }
        }

        /// <summary>
        /// Event logger
        /// </summary>
        /// <param name="msg"></param>
        private void LogEvent(string msg)
        {
            string header = "[" + DateTime.Now.ToString("yyyy-MM-dd HH:mm:ss") + "] ";
            EventsBuf += header + msg + "\n";
        }
        #endregion

        #region Commands
        public RelayCommand ConnectCommand { get; private set; }
        public RelayCommand BuffersClearCommand { get; private set; }

        private void CommandsBind()
        {
            ConnectCommand = new RelayCommand(Connect_Execute);
            BuffersClearCommand = new RelayCommand(BuffersClear_Execute);
        }

        private void Connect_Execute(object p)
        {
            if (ConnectAction.Equals(txtConnect))
            {
                if (!string.IsNullOrEmpty(ServerIp))
                {
                    ClientConnect();
                }
            }
            else
            {
                ClientDisconnect();
            }
        }

        private void BuffersClear_Execute(object p)
        {
            BuffersClear();
        }
        #endregion

        #region INotifyPropertyChanged
        public event PropertyChangedEventHandler PropertyChanged;

        /// <summary>
        /// Raise PropertyChanged event
        /// </summary>
        /// <param name="propertyName">property name</param>
        protected virtual void OnPropertyChanged([System.Runtime.CompilerServices.CallerMemberName] string propertyName = null)
        {
            if (PropertyChanged != null)
            {
                PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
            }
        }

        /// <summary>
        /// Set a propetry that raises a PropertyChanged event
        /// </summary>
        /// <typeparam name="T">property type</typeparam>
        /// <param name="field">property reference</param>
        /// <param name="value">new value</param>
        /// <param name="propertyName">property name (auto-set)</param>
        /// <returns>true if property modified, false if unchanged</returns>
        protected bool SetField<T>(ref T field, T value,
            [System.Runtime.CompilerServices.CallerMemberName] string propertyName = null)
        {
            if (!EqualityComparer<T>.Default.Equals(field, value))
            {
                field = value;
                OnPropertyChanged(propertyName);
                return true;
            }
            else
            {
                return false;
            }
        }
        #endregion
    }
}
